#define _GNU_SOURCE
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include "uhumenu.h"

struct charset *first_charset = NULL;
struct language *first_language = NULL;

static const char *try_locales[] = {
	"af_ZA",
	"am_ET",
	"ar_AE", "ar_BH", "ar_DZ", "ar_EG", "ar_IN", "ar_IQ", "ar_JO",
	"ar_KW", "ar_LB", "ar_LY", "ar_MA", "ar_OM", "ar_QA", "ar_SA",
	"ar_SD", "ar_SY", "ar_TN", "ar_YE",
	"az_AZ",
	"be_BY",
	"bg_BG",
	"bn_BD", "bn_IN",
	"br_FR",
	"bs_BA",
	"ca_ES",
	"cs_CZ",
	"cy_GB",
	"da_DK",
	"de_DE", "de_AT", "de_BE", "de_CH", "de_LU",
	"el_GR",
	"en_US", "en_GB", "en_AU", "en_BW", "en_CA", "en_DK", "en_HK",
	"en_IE", "en_IN", "en_NZ", "en_PH", "en_SG", "en_ZA", "en_ZW",
	"es_ES", "es_AR", "es_BO", "es_CL", "es_CO", "es_CR", "es_DO",
	"es_EC", "es_GT", "es_HN", "es_MX", "es_NI", "es_PA", "es_PE",
	"es_PR", "es_PY", "es_SV", "es_US", "es_UY", "es_VE",
	"et_EE",
	"eu_ES",
	"fa_IR",
	"fi_FI",
	"fo_FO",
	"fr_FR", "fr_BE", "fr_CA", "fr_CH", "fr_LU",
	"ga_IE",
	"gd_GB",
	"gl_ES",
	"gv_GB",
	"he_IL",
	"hi_IN",
	"hr_HR",
	"hu_HU",
	"hy_AM",
	"id_ID",
	"is_IS",
	"it_IT", "it_CH",
	"iw_IL",
	"ja_JP",
	"ka_GE",
	"kl_GL",
	"ko_KR",
	"kw_GB",
	"lt_LT",
	"lug_UG",
	"lv_LV",
	"mi_NZ",
	"mk_MK",
	"ml_IN",
	"mr_IN",
	"ms_MY",
	"mt_MT",
	"nl_NL", "nl_BE",
	"nn_NO",
	"no_NO",
	"oc_FR",
	"pl_PL",
	"pt_PT", "pt_BR",
	"ro_RO",
	"ru_RU", "ru_UA",
	"se_NO",
	"sk_SK",
	"sl_SI",
	"sq_AL",
	"sr_YU",
	"sv_FI",
	"sv_SE",
	"ta_IN",
	"te_IN",
	"tg_TJ",
	"th_TH",
	"ti_ER", "ti_ET",
	"tl_PH",
	"tr_TR",
	"tt_RU",
	"uk_UA",
	"ur_PK",
	"uz_UZ",
	"vi_VN",
	"wa_BE",
	"yi_US",
	"zh_CN", "zh_HK", "zh_SG", "zh_TW",
	NULL
};

/*
 * newlocale2:
 * if name=it_IT, utf8=0 then try it_IT
 * if name=it, utf8=0 then try it, it_IT and it_CH in this order
 * if name=it, utf8=1 then try it.UTF-8, it_IT.UTF-8 and it_CH.UTF-8
 * if name=ru.KOI8-R, utf8=0 then try ru.KOI8-R, ru_RU.KOI8-R, ru_UA.KOI8-R
 * if name=ru.KOI8-R, utf8=1 then try ru.UTF-8, ru_RU.UTF-8, ru_UA.UTF-8
 * etc...
 */

locale_t newlocale2 (const char *name, int utf8)
{
	int i, l;
	locale_t new;
	char lang[32];
	char encoding[32];
	char buf[64];
	char *p;

	p = strchr(name, '.');
	if (p != NULL) {
		strncpy(lang, name, p - name);
		lang[p - name] = '\0';
		strcpy(encoding, p);
	} else {
		strcpy(lang, name);
		encoding[0] = '\0';
	}
	if (utf8) strcpy(encoding, ".UTF-8");

	sprintf(buf, "%s%s", lang, encoding);
	new = newlocale(LC_ALL_MASK, buf, NULL);
	if (new != NULL) return new;
	if (strchr(name, '_') != NULL) return NULL;
	l = strlen(lang);
	i = 0;
	while (try_locales[i] != NULL) {
		if (!strncmp(try_locales[i], lang, l) && try_locales[i][l] == '_') {
			sprintf(buf, "%s%s", try_locales[i], encoding);
			new = newlocale(LC_ALL_MASK, buf, NULL);
			if (new != NULL) return new;
		}
		i++;
	}
	return NULL;
}

struct charset *find_or_create_charset (const char *name)
{
	struct charset **cp;
	struct charset *c;
	cp = &first_charset;
	c = first_charset;
	while (c != NULL) {
		if (!strcmp(c->name, name)) return c;
		cp = &c->next;
		c = c->next;
	}
	c = malloc(sizeof(struct charset));
	c->name = strdup(name);
	c->from_utf8 = iconv_open(name, "UTF-8");
	c->to_utf8 = iconv_open("UTF-8", name);
	c->next = NULL;
	*cp = c;
	return c;
}

struct language *find_or_create_language (const char *name)
{
	struct language **lp;
	struct language *l;
	lp = &first_language;
	l = first_language;
	while (l != NULL) {
		if (!strcmp(l->name, name)) return l;
		lp = &l->next;
		l = l->next;
	}
	l = malloc(sizeof(struct language));
	l->name = strdup(name);
	if (name[0] == '\0') name = "C";
	l->locale = newlocale2(name, 0);
	l->locale_utf8 = newlocale2(name, 1);
	if (l->locale != NULL) l->charset = find_or_create_charset(nl_langinfo_l(CODESET, l->locale));
	else l->charset = NULL;
	l->next = NULL;
	*lp = l;
	return l;
}




char *convert (iconv_t ic, /* XXX const */ char *from)
{
	char *to, *retval;
	size_t fromlen, tolen;
	fromlen = strlen(from);
	tolen = 6 * fromlen + 1;
	to = malloc(tolen);
	retval = to;
	iconv(ic, NULL, NULL, NULL, NULL);
	iconv(ic, &from, &fromlen, &to, &tolen);
	iconv(ic, NULL, NULL, NULL, NULL);
	*to = '\0';
	return retval;
}



int section_add_value (struct section *section, const char *key,
  const char *lang, const char *value)
{
	struct key **kp = &section->keys;
	struct key *k = section->keys;
	struct lang **lp;
	struct lang *l;
	while (k != NULL) {
		if (!strcmp(k->name, key)) {
			l = k->langs;
			lp = &k->langs;
			while (l != NULL) {
				if (!strcmp(l->language->name, lang)) return -1;
				lp = &l->next;
				l = l->next;
			}
			l = malloc(sizeof(struct lang));
			l->language = find_or_create_language(lang);
			l->value = strdup(value);
			l->value_8bit = NULL;
			l->next = NULL;
			*lp = l;
			return 0;
		}
		kp = &k->next;
		k = k->next;
	}
	k = malloc(sizeof(struct key));
	l = malloc(sizeof(struct lang));
	l->language = find_or_create_language(lang);
	l->value = strdup(value);
	l->value_8bit = NULL;
	l->next = NULL;
	k->name = strdup(key);
	k->langs = l;
	k->next = NULL;
	*kp = k;
	return 0;
}

struct section *item_add_section (struct item *item, const char *section)
{
	struct section *s;
	struct section **sp;
	s = item->sections;
	sp = &item->sections;
	while (s != NULL) {
		if (!strcmp(s->name, section)) {
			return s;
		}
		sp = &s->next;
		s = s->next;
	}
	s = malloc(sizeof(struct section));
	s->name = strdup(section);
	s->keys = NULL;
	s->next = NULL;
	*sp = s;
	return s;
}

static struct lang *key_find_exact_lang (const struct key *key, const char *lang)
{
	struct lang *l;

	if (lang == NULL) lang = "";
	l = key->langs;
	while (l != NULL) {
		if (!strcmp(l->language->name, lang)) return l;
		l = l->next;
	}
	return NULL;
}

struct lang *key_find_lang (const struct key *key, const char *lang)
{
	struct lang *l;
	char *p;
	char buf[16];

	if (lang == NULL) lang = "";
	strcpy(buf, lang);
	p = strchr(buf, '_');
	if (p != NULL) {
		l = key_find_exact_lang(key, buf);
		if (l != NULL) return l;
		*p = '\0';
	}
	if (buf[0] != '\0') {
		l = key_find_exact_lang(key, buf);
		if (l != NULL) return l;
	}
	return key_find_exact_lang(key, "");
}

struct key *section_find_key (const struct section *section, const char *key)
{
	struct key *k;
	k = section->keys;
	while (k != NULL) {
		if (!strcmp(k->name, key)) return k;
		k = k->next;
	}
	return NULL;
}

struct section *item_find_section (const struct item *item, const char *section)
{
	struct section *s;
	if (section == NULL || section[0] == '\0') section = "Desktop Entry";
	s = item->sections;
	while (s != NULL) {
		if (!strcmp(s->name, section)) return s;
		s = s->next;
	}
	return NULL;
}

struct key *item_find_key (const struct item *item, const char *section,
  const char *key)
{
	struct section *s;
	s = item_find_section(item, section);
	if (s == NULL) return NULL;
	return section_find_key(s, key);
}

char *get_value (const struct item *item, const char *section,
  const char *key, const char *lang)
{
	struct section *s;
	struct key *k;
	struct lang *l;

	s = item_find_section(item, section);
	if (s == NULL) return NULL;
	k = section_find_key(s, key);
	if (k == NULL) return NULL;
	l = key_find_lang(k, lang);
	if (l == NULL) return NULL;
	return l->value;
}

char *get_value_8bit (const struct item *item, const char *section,
  const char *key, const char *lang)
{
	struct section *s;
	struct key *k;
	struct lang *l;

	s = item_find_section(item, section);
	if (s == NULL) return NULL;
	k = section_find_key(s, key);
	if (k == NULL) return NULL;
	l = key_find_lang(k, lang);
	if (l == NULL) return NULL;
	return l->value_8bit;
}

int empty_dir (const char *path)
{
	int fd;
	DIR *d;
	struct dirent *de;
	struct stat st;
	int error = 0;

	fd = open(".", O_RDONLY);
	if (fd < 0) {
		return -1;
	}
	if (chdir(path) < 0) {
		close(fd);
		return -1;
	}
	if ((d = opendir(".")) == NULL) {
		fchdir(fd);
		close(fd);
		return -1;
	}
	while ((de = readdir(d)) != NULL) {
		if (!strcmp(de->d_name, "." )) continue;
		if (!strcmp(de->d_name, "..")) continue;
		if (lstat(de->d_name, &st) < 0) {
			error = 1;
			continue;
		}
		if (S_ISDIR(st.st_mode)) {
			if (empty_dir(de->d_name) < 0) error = 1;
			if (rmdir(de->d_name) < 0) error = 1;
		} else {
			if (unlink(de->d_name) < 0) error = 1;
		}
	}
	closedir(d);
	fchdir(fd);
	close(fd);
	if (error) return -1;
	return 0;
}
